// -*-c++-*-

/*!
	\file feditor_canvas.cpp
	\brief formation editor canvas class Source File.
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.	If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

// for compliers supporting precompling
#include <wx/wxprec.h>

#ifdef __BORLANDC__
#pragma hdrstop
#endif

// for compliers NOT supporting precompling
#ifndef WX_PRECOMP
#include <wx/wx.h>
#endif

#include <iostream>
#include <cassert>
#include <algorithm>

#include "id.h"
#include "app_config.h"
#include "gdi_config.h"
#include "feditor_frame.h"
#include "feditor_data.h"

#include "feditor_canvas.h"

/*-------------------------------------------------------------------*/
/*!

*/
FEditorCanvas::FEditorCanvas( FEditorFrame * parent )
    : wxWindow( parent, -1, wxDefaultPosition, wxDefaultSize,
                wxSIMPLE_BORDER /*| wxNO_FULL_REPAINT_ON_RESIZE*/ )
    , M_parent( parent )
    , M_scale( 1.0 )
    , M_center( 0, 0 )
    , M_field_brush( GDIConfig::FIELD_COLOR, wxSOLID )
    , M_field_dark_brush( GDIConfig::FIELD_DARK_COLOR, wxSOLID )
    , M_line_pen( GDIConfig::LINE_COLOR, 1, wxSOLID )
    , M_triangle_pen( wxColor( 255, 127, 0 ), 1, wxSOLID )
    , M_triangle_font( 10, wxDEFAULT, wxNORMAL, wxNORMAL, false, wxT( "" ), wxFONTENCODING_ISO8859_1 ) //wxFONTENCODING_SYSTEM )
    , M_area_pen( wxColor( 127, 127, 127 ), 1, wxSOLID )
    , M_ball_pen( GDIConfig::BALL_PEN_COLOR, 1, wxSOLID )
    , M_ball_brush( GDIConfig::BALL_COLOR, wxSOLID )
    , M_player_pen( GDIConfig::PLAYER_PEN_COLOR, 1, wxSOLID )
    , M_select_pen( *wxWHITE, 2, wxSOLID )
    , M_left_team_brush( GDIConfig::LEFT_TEAM_COLOR, wxSOLID )
    , M_right_team_brush( GDIConfig::RIGHT_TEAM_COLOR, wxSOLID )
    , M_mirror_brush( GDIConfig::DEBUG_TEAMMATE_COLOR, wxSOLID )
    , M_player_font( 10, wxDEFAULT, wxNORMAL, wxNORMAL, false, wxT( "" ), wxFONTENCODING_ISO8859_1 ) //wxFONTENCODING_SYSTEM )
{
    assert( parent );

    //parent->SetClientSize( static_cast< int >( pitch_len / M_scale ),
    //static_cast< int >( pitch_width / M_scale ) );

    createWindows();
    connectEvents();

    M_canvas_bmp.Create( this->GetClientSize().GetWidth(),
                         this->GetClientSize().GetHeight() );

    updateSize();
}

/*-------------------------------------------------------------------*/
/*!
  destructor
*/
FEditorCanvas::~FEditorCanvas()
{
    //std::cerr << "delete FEditorCanvas" << std::endl;
}

/*-------------------------------------------------------------------*/
/*!

*/
void
FEditorCanvas::connectEvents()
{
    //////////////////////////////////////////////////////////////
    // paint event
    Connect( wxID_ANY, wxEVT_PAINT,
             wxPaintEventHandler( FEditorCanvas::handlePaint ) );
    // mouse event
    Connect( wxID_ANY, wxEVT_LEFT_DOWN,
             wxMouseEventHandler( FEditorCanvas::handleLeftDown ) );
    Connect( wxID_ANY, wxEVT_LEFT_UP,
             wxMouseEventHandler( FEditorCanvas::handleLeftUp ) );
    Connect( wxID_ANY, wxEVT_MIDDLE_DOWN,
             wxMouseEventHandler( FEditorCanvas::handleMiddleDown ) );
    Connect( wxID_ANY, wxEVT_MIDDLE_UP,
             wxMouseEventHandler( FEditorCanvas::handleMiddleUp ) );
    Connect( wxID_ANY, wxEVT_RIGHT_DOWN,
             wxMouseEventHandler( FEditorCanvas::handleRightDown ) );
    Connect( wxID_ANY, wxEVT_RIGHT_UP,
             wxMouseEventHandler( FEditorCanvas::handleRightUp ) );
    Connect( wxID_ANY, wxEVT_MOTION,
             wxMouseEventHandler( FEditorCanvas::handleMouseMotion ) );
}


/*-------------------------------------------------------------------*/
/*!

*/
void
FEditorCanvas::createWindows()
{

}

/*-------------------------------------------------------------------*/
/*!

*/
void
FEditorCanvas::updateSize()
{
    M_center.x = this->GetClientSize().x / 2;
    M_center.y = this->GetClientSize().y / 2;

    const double pitch_l = ( AppConfig::PITCH_LENGTH
                             + AppConfig::PITCH_MARGIN * 2.0 );
    const double pitch_w = ( AppConfig::PITCH_WIDTH
                             + AppConfig::PITCH_MARGIN * 2.0 );

    M_scale = this->GetClientSize().x / pitch_l;
    if ( pitch_w * M_scale > this->GetSize().y )
    {
        M_scale = this->GetClientSize().y / pitch_w;
    }
    M_scale = std::max( 2.0, M_scale );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
FEditorCanvas::handlePaint( wxPaintEvent & WXUNUSED( event ) )
{
    wxPaintDC dc( this );

    draw( dc );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
FEditorCanvas::draw()
{
    wxClientDC dc( this );

    draw( dc );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
FEditorCanvas::draw( wxDC & dc )
{
    // check canvas size
    const wxSize cur_size = this->GetClientSize();
    if ( M_canvas_bmp.GetWidth() != cur_size.GetWidth()
         || M_canvas_bmp.GetHeight() != cur_size.GetHeight() )
    {
        M_canvas_bmp.Create( cur_size.GetWidth(), cur_size.GetHeight() );
        updateSize();
    }

    // create buffer DC
    wxMemoryDC mem_dc;
    mem_dc.SelectObject( M_canvas_bmp );

    drawField( mem_dc );
    drawTrainingData( mem_dc );
    drawPlayers( mem_dc );
    drawBall( mem_dc );

    // blit buffer DC to 'dc'
    dc.Blit( 0, 0,
             cur_size.GetWidth(), cur_size.GetHeight(),
             &mem_dc, 0, 0 );

    mem_dc.SelectObject( wxNullBitmap );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
FEditorCanvas::drawField( wxDC & dc )
{
    //--------------------------------------
    // fill background
    dc.SetBackground( M_field_brush );
    dc.Clear();

    //--------------------------------------
    drawContainedArea( dc );

    // set screen coordinates of field
    const wxCoord left_x   = getScreenX( - AppConfig::PITCH_LENGTH * 0.5 );
    const wxCoord right_x  = getScreenX( + AppConfig::PITCH_LENGTH * 0.5 );
    const wxCoord top_y    = getScreenY( - AppConfig::PITCH_WIDTH * 0.5 );
    const wxCoord bottom_y = getScreenY( + AppConfig::PITCH_WIDTH * 0.5 );

    // fill bottom side
    //dc.SetPen( *wxTRANSPARENT_PEN );
    //dc.SetBrush( M_field_dark_brush );
    //dc.DrawRectangle( left_x, M_center.y,
    //right_x - left_x, (bottom_y - top_y)/2 );

    //--------------------------------------
    // draw lines
    dc.SetPen( M_line_pen );
    dc.SetBrush( *wxTRANSPARENT_BRUSH );
    // side lines & goal lines
    dc.DrawLine( left_x, top_y, right_x, top_y );
    dc.DrawLine( right_x, top_y, right_x, bottom_y );
    dc.DrawLine( right_x, bottom_y, left_x, bottom_y );
    dc.DrawLine( left_x, bottom_y, left_x, top_y );
    // center line
    dc.DrawLine( M_center.x, top_y, M_center.x, bottom_y );
    // center circle
    const wxCoord center_radius = scale( AppConfig::CENTER_CIRCLE_R );
    dc.DrawCircle( M_center, center_radius );
    // draw penalty area box
    const wxCoord pen_top_y    = getScreenY( - AppConfig::PENALTY_AREA_WIDTH / 2 );
    const wxCoord pen_bottom_y = getScreenY( + AppConfig::PENALTY_AREA_WIDTH / 2 );
    // left penalty box
    wxCoord pen_x = getScreenX( -( AppConfig::PITCH_LENGTH*0.5
                                   - AppConfig::PENALTY_AREA_LENGTH ) );
    dc.DrawLine( left_x, pen_top_y, pen_x, pen_top_y );
    dc.DrawLine( pen_x, pen_top_y, pen_x, pen_bottom_y );
    dc.DrawLine( pen_x, pen_bottom_y, left_x, pen_bottom_y );
    // right penalty box
    pen_x = getScreenX( +( AppConfig::PITCH_LENGTH*0.5
                           - AppConfig::PENALTY_AREA_LENGTH ) );
    dc.DrawLine( right_x, pen_top_y, pen_x, pen_top_y );
    dc.DrawLine( pen_x, pen_top_y, pen_x, pen_bottom_y );
    dc.DrawLine( pen_x, pen_bottom_y, right_x, pen_bottom_y );
    // draw goal area box
    const wxCoord goal_area_y_abs = scale( AppConfig::GOAL_AREA_WIDTH*0.5 );
    const wxCoord goal_area_top_y = M_center.y - goal_area_y_abs;
    const wxCoord goal_area_bottom_y = M_center.y + goal_area_y_abs;
    // left goal area
    wxCoord goal_area_x = getScreenX( - AppConfig::PITCH_LENGTH*0.5
                                      + AppConfig::GOAL_AREA_LENGTH );
    dc.DrawLine( left_x, goal_area_top_y, goal_area_x, goal_area_top_y );
    dc.DrawLine( goal_area_x, goal_area_top_y, goal_area_x, goal_area_bottom_y );
    dc.DrawLine( goal_area_x, goal_area_bottom_y, left_x, goal_area_bottom_y );
    // right goal area
    goal_area_x = getScreenX( AppConfig::PITCH_LENGTH*0.5
                              - AppConfig::GOAL_AREA_LENGTH );
    dc.DrawLine( right_x, goal_area_top_y, goal_area_x, goal_area_top_y );
    dc.DrawLine( goal_area_x, goal_area_top_y, goal_area_x, goal_area_bottom_y );
    dc.DrawLine( goal_area_x, goal_area_bottom_y, right_x, goal_area_bottom_y );

    //--------------------------------------
    // draw goal boxes
    dc.SetPen( *wxBLACK_PEN );
    dc.SetBrush( *wxBLACK_BRUSH );
    const wxCoord goal_top_y = getScreenY( - AppConfig::GOAL_WIDTH*0.5 );
    const wxCoord goal_size_x = scale( AppConfig::GOAL_DEPTH );
    const wxCoord goal_size_y = scale( AppConfig::GOAL_WIDTH );
    // left goal
    dc.DrawRectangle( getScreenX( - AppConfig::PITCH_LENGTH*0.5
                                  - AppConfig::GOAL_DEPTH ),
                      goal_top_y, goal_size_x, goal_size_y );
    // right goal
    dc.DrawRectangle( getScreenX( AppConfig::PITCH_LENGTH*0.5 ) + 1,
                      goal_top_y, goal_size_x + 1, goal_size_y );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
FEditorCanvas::drawContainedArea( wxDC & dc )
{
    boost::shared_ptr< FEditorData > ptr = M_data_ptr.lock();
    if ( ! ptr )
    {
        return ;
    }

    boost::shared_ptr< const Formation > f = ptr->formation();
    if ( ! f )
    {
        return;
    }

    if ( f->getMethodName() != "Delaunay" )
    {
        return;
    }

    if ( ptr->getTriangulation().vertices().size() < 3 )
    {
        return;
    }

    // draw polygon
    const rcsc::DelaunayTriangulation::TrianglePtr tri
        = ptr->getTriangulation().findTriangleContains( ptr->getBall() );
    if ( tri )
    {
        dc.SetPen( *wxTRANSPARENT_PEN );
        dc.SetBrush( M_field_dark_brush );
        wxPoint vertices[3] = { wxPoint( getScreenX( tri->vertex( 0 )->pos().x ),
                                         getScreenY( tri->vertex( 0 )->pos().y ) ),
                                wxPoint( getScreenX( tri->vertex( 1 )->pos().x ),
                                         getScreenY( tri->vertex( 1 )->pos().y ) ),
                                wxPoint( getScreenX( tri->vertex( 2 )->pos().x ),
                                         getScreenY( tri->vertex( 2 )->pos().y ) ),
        };

        dc.DrawPolygon( 3, vertices );
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
FEditorCanvas::drawTrainingData( wxDC & dc )
{
    boost::shared_ptr< FEditorData > ptr = M_data_ptr.lock();
    if ( ! ptr )
    {
        return;
    }

    boost::shared_ptr< const Formation > f = ptr->formation();
    if ( ! f )
    {
        return;
    }

    if (  f->getMethodName() != "Delaunay" )
    {
        dc.SetPen( M_triangle_pen );
        dc.SetBrush( *wxTRANSPARENT_BRUSH );

        const std::list< Formation::Snapshot >::const_iterator train_end
            = ptr->getTrainingData().end();
        for ( std::list< Formation::Snapshot >::const_iterator train_it
                  = ptr->getTrainingData().begin();
              train_it != train_end;
              ++train_it )
        {
            dc.DrawCircle( getScreenX( train_it->ball_.x ),
                           getScreenY( train_it->ball_.y ),
                           scale( 0.8 ) );
        }

        return;
    }


    if ( ptr->getTriangulation().vertices().size() < 3 )
    {
        // too few kernel points
        // no valid triangulation
        return;
    }

    dc.SetPen( M_triangle_pen );
    dc.SetBrush( *wxTRANSPARENT_BRUSH );

    const std::map< int, rcsc::DelaunayTriangulation::EdgePtr >::const_iterator end
        = ptr->getTriangulation().edgeMap().end();
    for ( std::map< int, rcsc::DelaunayTriangulation::EdgePtr >::const_iterator it
              = ptr->getTriangulation().edgeMap().begin();
          it != end;
          ++it )
    {
        dc.DrawLine( getScreenX( it->second->vertex( 0 )->pos().x ),
                     getScreenY( it->second->vertex( 0 )->pos().y ),
                     getScreenX( it->second->vertex( 1 )->pos().x ),
                     getScreenY( it->second->vertex( 1 )->pos().y ) );
    }

    dc.SetFont( M_player_font );
    dc.SetTextForeground( *wxRED );
    dc.SetBackgroundMode( wxTRANSPARENT );

    const int radius = scale( 0.7 );
    int count = 0;
    const std::list< Formation::Snapshot >::const_iterator v_end
        = ptr->getTrainingData().end();
    for ( std::list< Formation::Snapshot >::const_iterator it
              = ptr->getTrainingData().begin();
          it != v_end;
          ++it )
    {
        int ix = getScreenX( it->ball_.x );
        int iy = getScreenY( it->ball_.y );
        wxString str;
        str.sprintf( wxT( "%2d" ), ++count );
        dc.DrawText( str,
                     ix + radius, iy - radius );
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
FEditorCanvas::drawPlayers( wxDC & dc )
{
    boost::shared_ptr< FEditorData > ptr = M_data_ptr.lock();
    if ( ! ptr )
    {
        return;
    }

    boost::shared_ptr< const Formation > f = ptr->formation();
    if ( ! f )
    {
        return;
    }

    const int radius = scale( 1.2 );

    ////////////////////////////////////////////////////////////
    dc.SetPen( M_player_pen );
    dc.SetFont( M_player_font );
    dc.SetTextForeground( *wxWHITE );
    dc.SetBackgroundMode( wxTRANSPARENT );
    ////////////////////////////////////////////////////////////
    // draw right players
#if 0
    {
        dc.SetBrush( M_right_team_brush );

        const std::vector< rcsc::Vector2D >::const_iterator
            selection = ( ptr->getSelectType() == FEditorData::SELECT_OPP
                          ? ( ptr->getOppPlayersPos().begin()
                              + ptr->getSelectIndex() )
                          : ptr->getOppPlayersPos().end() );
        int unum = 1;
        for ( std::vector< rcsc::Vector2D >::const_iterator
                  it = ptr->getOppPlayersPos().begin();
              it != ptr->getOppPlayersPos().end();
              ++it, ++unum )
        {
            int ix = getScreenX( it->x );
            int iy = getScreenY( it->y );
            if ( selection == it )
            {
                dc.SetPen( M_select_pen );
                dc.DrawCircle( ix, iy, radius + 3 );
                dc.SetPen( M_player_pen );
            }
            else
            {
                dc.DrawCircle( ix, iy, radius );
            }
            wxString player_str;
            player_str.Printf( wxT( "%2d" ), unum );
            dc.DrawText( player_str,
                         ix + radius, iy - radius );

        }
    }
#endif
    ////////////////////////////////////////////////////////////
    // draw left players
    {
        const std::vector< rcsc::Vector2D >::const_iterator
            selection = ( ptr->getSelectType() == FEditorData::SELECT_OUR
                          ? ( ptr->getOurPlayers().begin()
                              + ptr->getSelectIndex() )
                          : ptr->getOurPlayers().end() );
        int unum = 1;
        for ( std::vector< rcsc::Vector2D >::const_iterator
                  it = ptr->getOurPlayers().begin();
              it != ptr->getOurPlayers().end();
              ++it, ++unum )
        {
            int ix = getScreenX( it->x );
            int iy = getScreenY( it->y );
            if ( f->isMirrorType( unum ) )
            {
                dc.SetBrush( M_mirror_brush );
                dc.DrawCircle( ix, iy, radius );
            }
            else if ( selection == it )
            {
                dc.SetPen( M_select_pen );
                dc.SetBrush( M_left_team_brush );
                dc.DrawCircle( ix, iy, radius + 3 );
                dc.SetPen( M_player_pen );
            }
            else
            {
                dc.SetBrush( M_left_team_brush );
                dc.DrawCircle( ix, iy, radius );
            }
            wxString player_str;
            player_str.Printf( wxT( "%2d" ), unum );
            dc.DrawText( player_str,
                         ix + radius, iy - radius );
        }
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
FEditorCanvas::drawBall( wxDC & dc )
{
    boost::shared_ptr< FEditorData > ptr = M_data_ptr.lock();
    if ( ! ptr )
    {
        return;
    }

    dc.SetPen( M_ball_pen );

    if ( ptr->isBallDraggable() )
    {
        dc.SetBrush( M_ball_brush );
    }
    else
    {
        dc.SetBrush( *wxLIGHT_GREY_BRUSH );
    }

    int radius = ( ptr->getSelectType() == FEditorData::SELECT_BALL
                   ? scale( 1.0 )
                   : scale( 0.7 ) );

    dc.DrawCircle( getScreenX( ptr->getBall().x ),
                   getScreenY( ptr->getBall().y ),
                   radius );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
FEditorCanvas::handleLeftDown( wxMouseEvent & event )
{
    M_mouse_state[0].down( event.GetPosition() );

    wxPoint point = event.GetPosition();
    boost::shared_ptr< FEditorData > ptr = M_data_ptr.lock();
    if ( ptr )
    {
        ptr->selectObject( ( point.x - M_center.x ) / M_scale,
                           ( point.y - M_center.y ) / M_scale );
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
FEditorCanvas::handleLeftUp( wxMouseEvent & event )
{
    M_mouse_state[0].up();

    boost::shared_ptr< FEditorData > ptr = M_data_ptr.lock();
    if ( ptr )
    {
        ptr->releaseObject( ( event.GetPosition().x - M_center.x ) / M_scale,
                            ( event.GetPosition().y - M_center.y ) / M_scale );
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
FEditorCanvas::handleMiddleDown( wxMouseEvent & event )
{
    M_mouse_state[1].down( event.GetPosition() );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
FEditorCanvas::handleMiddleUp( wxMouseEvent & event )
{
    M_mouse_state[1].up();
}

/*-------------------------------------------------------------------*/
/*!

*/
void
FEditorCanvas::handleRightDown( wxMouseEvent & event )
{
    M_mouse_state[2].down( event.GetPosition() );

    if ( boost::shared_ptr< FEditorData > ptr = M_data_ptr.lock() )
    {
        if ( ptr->isBallDraggable() )
        {
            ptr->dropBallTo( ( event.GetPosition().x - M_center.x ) / M_scale,
                             ( event.GetPosition().y - M_center.y ) / M_scale );
        }
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
FEditorCanvas::handleRightUp( wxMouseEvent & event )
{
    M_mouse_state[2].up();
}

/*-------------------------------------------------------------------*/
/*!

*/
void
FEditorCanvas::handleMouseMotion( wxMouseEvent & event )
{
    wxPoint point = event.GetPosition();
    for ( int i = 0; i < 3; ++i )
    {
        M_mouse_state[i].motion( point );
    }
    M_parent->updateStatusText( ( point.x - M_center.x ) / M_scale,
                                ( point.y - M_center.y ) / M_scale );

    if ( M_mouse_state[0].isDragged() )
    {
        boost::shared_ptr< FEditorData > ptr = M_data_ptr.lock();
        if ( ptr
             && ptr->getSelectType() != FEditorData::NO_SELECT )
        {
            ptr->setObjectDragPoint( ( point.x - M_center.x ) / M_scale,
                                     ( point.y - M_center.y ) / M_scale );
        }
    }

}
